##
# File:    OeMoleculeFactory.py
# Author:  jdw
# Date:    2-Oct-2019
# Version: 0.001
#
# Updates:
# 2-Oct-2019  jdw adapted from OeBuildMol()
##
"""
Classes to build OE molecule objects from chemical component definition data.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import hashlib
import logging
from collections import defaultdict, namedtuple

from openeye import oechem
from openeye import oegraphsim
from openeye import oequacpac

from rcsb.utils.chem.PdbxChemComp import PdbxChemCompAtomIt, PdbxChemCompBondIt, PdbxChemCompDescriptorIt, PdbxChemCompIt
from rcsb.utils.chem.PdbxChemCompConstants import PdbxChemCompConstants


logger = logging.getLogger(__name__)

ComponentDetails = namedtuple("ComponentDetails", "ccId formula ifCharge")
ComponentAtom = namedtuple("ComponentAtom", "name aType isAromatic isChiral CIP fCharge")
ComponentBond = namedtuple("ComponentBond", "iType isAromatic CIP")
ComponentDescriptors = namedtuple("ComponentDescriptors", "smiles isoSmiles inchi inchiKey")
DescriptorInstance = namedtuple("DescriptorInstance", "type program")


class OeMoleculeFactory(object):
    """ Utility methods for constructing OEGraphMols from chemical component definition objects.
    """

    def __init__(self, verbose=False):
        self.__verbose = verbose
        self.__oeErrorLevel = oechem.OEErrorLevel_Info
        self.__ccId = None
        self.__oeMol = None
        #
        # Source data categories objects from chemical component definitions.
        self.__dataContainer = None
        #
        self.__descriptorBuildType = None
        self.__descriptor = None
        #
        self.__isDefinitionSet = False
        #

    def __oeClear(self):
        #
        self.__oeMol = None
        self.__descriptorBuildType = None
        self.__descriptor = None

    def setQuiet(self):
        """Suppress OE warnings and processing errors
        """
        oechem.OEThrow.SetLevel(oechem.OEErrorLevel_Quiet)
        self.__oeErrorLevel = oechem.OEErrorLevel_Quiet

    def setDebug(self, flag):
        self.__verbose = flag

    def setChemCompDef(self, dataContainer):
        try:
            if dataContainer.exists("chem_comp"):
                ccIt = PdbxChemCompIt(dataContainer)
                for cc in ccIt:
                    ccId = cc.getId()
                self.__ccId = ccId
            else:
                self.__ccId = dataContainer.getName()

            self.__dataContainer = dataContainer
            if self.__dataContainer.exists("chem_comp_atom"):
                self.__isDefinitionSet = True
                return self.__ccId
            else:
                return None
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return None

    def setOeMol(self, inpOeMol, ccId):
        """  Load this object with an existing oeMOL()
        """
        self.__oeClear()
        self.__oeMol = oechem.OEGraphMol(inpOeMol)
        self.__ccId = ccId
        # self.getElementCounts()
        return ccId

    def setDescriptor(self, descriptor, molBuildType, ccId):
        self.__oeClear()
        self.__ccId = ccId
        self.__descriptorBuildType = molBuildType
        self.__descriptor = descriptor
        return ccId

    def addFingerPrints(self, fpTypeList):
        """ Add fingerprints (fpType) as data sections in the current molecule.

        Args:
            fpTypeList (list): list of fingerprint types [TREE,PATH,MACCS,CIRCULAR,LINGO].

        Returns:
            bool: True for success or False otherwise
        """
        if not self.__oeMol:
            return False
        myFpTypeList = fpTypeList if fpTypeList else []
        fpD = {
            "TREE": oegraphsim.OEFPType_Tree,
            "CIRCULAR": oegraphsim.OEFPType_Circular,
            "PATH": oegraphsim.OEFPType_Path,
            "MACCS": oegraphsim.OEFPType_MACCS166,
            "LINGO": oegraphsim.OEFPType_Lingo,
        }
        ret = True
        for fpType in myFpTypeList:
            if fpType in fpD:
                tag = "FP_" + fpType
                fp = oegraphsim.OEFingerPrint()
                ok = oegraphsim.OEMakeFP(fp, self.__oeMol, fpD[fpType])
                if ok and fp.IsValid():
                    self.__oeMol.SetData(tag, fp)
                    fpOk = True
                ret = ret and fpOk
        return ret

    def updateOePerceptions3D(self, oeMol, aromaticModel=oechem.OEAroModelOpenEye):
        oeMol.SetDimension(3)
        oechem.OEFindRingAtomsAndBonds(oeMol)
        oechem.OEPerceiveChiral(oeMol)
        #
        # Other aromatic models: oechem.OEAroModelMDL or OEAroModelDaylight or oechem.OEAroModelOpenEye
        #
        if aromaticModel is not None:
            oechem.OEAssignAromaticFlags(oeMol, aromaticModel)
        oechem.OE3DToInternalStereo(oeMol)
        return oeMol

    def updateOePerceptions2D(self, oeMol, aromaticModel=oechem.OEAroModelOpenEye):
        #
        # run standard perceptions --
        oechem.OEFindRingAtomsAndBonds(oeMol)
        oechem.OEPerceiveChiral(oeMol)
        #
        # Other aromatic models: oechem.OEAroModelMDL or OEAroModelDaylight or oechem.OEAroModelOpenEye
        #
        if aromaticModel is not None:
            oechem.OEAssignAromaticFlags(oeMol, aromaticModel)
        return oeMol

    def __transferAromaticFlagsOE(self, oeMol):
        """Manually assign/transfer aromatic flags from chemical component definition.

           Must follow other perceptions -
        """
        atIt = PdbxChemCompAtomIt(self.__dataContainer)
        ccAtomD = {}
        for at in atIt:
            atName = at.getName()
            isAromatic = at.isAromatic()
            ccAtomD[atName] = isAromatic
        ccBondD = {}
        bndIt = PdbxChemCompBondIt(self.__dataContainer)
        for bnd in bndIt:
            atIdI, atIdJ = bnd.getBond()
            isAromatic = bnd.isAromatic()
            ccBondD[(atIdI, atIdJ)] = isAromatic
            ccBondD[(atIdJ, atIdI)] = isAromatic
        #
        for atom in oeMol.GetAtoms():
            atName = atom.GetName()
            isAromatic = ccAtomD[atName]
            atom.SetAromatic(isAromatic)

        for bond in oeMol.GetBonds():
            atNameI = bond.GetBgn().GetName()
            atNameJ = bond.GetEnd().GetName()
            isAromatic = ccBondD[(atNameI, atNameJ)]
            bond.SetAromatic(isAromatic)
        #
        return oeMol

    def updateCIPStereoOE(self, oeMol):
        """ Manually assign OE CIP stereo perceptions for each atom and bond.

            Redundant of oechem.OE3DToInternalStereo(oeMol) -
        """
        for atom in oeMol.GetAtoms():
            cipV = oechem.OEPerceiveCIPStereo(oeMol, atom)
            if cipV in [oechem.OECIPAtomStereo_S, oechem.OECIPAtomStereo_R]:
                logger.info("%s CIP stereo is %r", atom.GetName(), cipV)
                oechem.OESetCIPStereo(oeMol, atom, cipV)
                cipV2 = oechem.OEPerceiveCIPStereo(oeMol, atom)
                logger.info("%s perceive after set CIP stereo (%r) is %r", atom.GetName(), atom.HasStereoSpecified(), cipV2)

        #
        for bond in oeMol.GetBonds():
            if bond.GetOrder() == 2:
                cipV = oechem.OEPerceiveCIPStereo(oeMol, bond)
                if cipV in [oechem.OECIPBondStereo_Z, oechem.OECIPBondStereo_E]:
                    logger.info("%s-%s CIP stereo is %r", bond.GetBgn().GetName(), bond.GetEnd().GetName(), cipV)
                    oechem.OESetCIPStereo(oeMol, bond, cipV)
                    cipV2 = oechem.OEPerceiveCIPStereo(oeMol, bond)
                    logger.info("%s-%s perceive after set CIP stereo (%r) is %r", bond.GetBgn().GetName(), bond.GetEnd().GetName(), bond.HasStereoSpecified(), cipV2)
        return oeMol

    def getGraphMolSuppressH(self):
        """ Return the current constructed OE molecule with hydrogens suppressed.
        """
        # OESuppressHydrogens(self.__oeMol, retainPolar=False,retainStereo=True,retainIsotope=True)
        oechem.OESuppressHydrogens(self.__oeMol)
        return self.__oeMol

    def getMol(self):
        """ Return a copy of the current constructed OE molecule.
        """
        return oechem.OEMol(self.__oeMol) if self.__oeMol else None

    def getGraphMol(self):
        """ Return a copy of current constructed Graph OE molecule.
        """
        return oechem.OEGraphMol(self.__oeMol) if self.__oeMol else None

    def getCanSMILES(self):
        """ Return the cannonical SMILES string derived from the current OD molecule.
        """
        return oechem.OECreateCanSmiString(self.__oeMol) if self.__oeMol else None

    def getIsoSMILES(self):
        """ Return the cannonical stereo SMILES string derived from the current OE molecule.
        """
        return oechem.OECreateIsoSmiString(self.__oeMol) if self.__oeMol else None

    def getFormula(self):
        """ Return the Hill order formulat  derived from the current OE molecule.
        """
        return oechem.OEMolecularFormula(self.__oeMol) if self.__oeMol else None

    def getFormalCharge(self):
        return oechem.OENetCharge(self.__oeMol) if self.__oeMol else 0

    def getInChIKey(self):
        """ Return the InChI key derived from the current OE molecule.
        """
        return oechem.OECreateInChIKey(self.__oeMol) if self.__oeMol else None

    def getInChI(self):
        """ Return the InChI string derived from the current OE molecule.
        """
        return oechem.OECreateInChI(self.__oeMol) if self.__oeMol else None

    def getTitle(self):
        """ Return the title assigned to the current OE molecule
        """
        return self.__oeMol.GetTitle() if self.__oeMol else None

    def getCcId(self):
        """ Return the CC id of this object -
        """
        return self.__ccId

    def addSdTags(self):
        smi = self.getIsoSMILES()
        if smi:
            oechem.OESetSDData(self.__oeMol, "OPENEYE_ISO_SMILES", smi)
        inchi = self.getInChI()
        if inchi:
            oechem.OESetSDData(self.__oeMol, "OPENEYE_INCHI", inchi)
        inchiKey = self.getInChIKey()
        if inchiKey:
            oechem.OESetSDData(self.__oeMol, "OPENEYE_INCHIKEY", inchiKey)
        formula = self.getFormula()
        if formula:
            oechem.OESetSDData(self.__oeMol, "FORMULA", formula)

    def setSimpleAtomNames(self):
        """
        """
        if self.__oeMol:
            for atom in self.__oeMol.GetAtoms():
                atom.SetIntType(atom.GetAtomicNum())
                atom.SetType(oechem.OEGetAtomicSymbol(atom.GetAtomicNum()))
            oechem.OETriposAtomNames(self.__oeMol)

    def getElementCounts(self, addExplicitHydrogens=False, useSymbol=False):
        """ Get the dictionary of element counts (eg. eD[iAtNo]=iCount).
        """
        eD = {}
        if self.__oeMol:
            if addExplicitHydrogens:
                if not oechem.OEAddExplicitHydrogens(self.__oeMol):
                    logger.warning("%s explict hydrogen addition fails", self.__ccId)
            # calculate from current oeMol
            try:
                eD = {}
                for atom in self.__oeMol.GetAtoms():
                    atNo = atom.GetAtomicNum()
                    if atNo not in eD:
                        eD[atNo] = 1
                    else:
                        eD[atNo] += 1
            except Exception:
                pass
        rD = {PdbxChemCompConstants.periodicTable[atNo - 1]: v for atNo, v in eD.items()} if useSymbol else eD
        #
        return rD

    def getOeMoleculeFeatures(self, filterHydrogens=False):
        """Get the essential features of the constructed OEMol for the input component.
        """
        formula = self.getFormula()
        # ccId = self.__ccId
        ccId = self.getTitle()
        ifCharge = oechem.OENetCharge(self.__oeMol)
        inchi = self.getInChI()
        inchiKey = self.getInChIKey()
        smiles = self.getCanSMILES()
        isoSmiles = self.getIsoSMILES()
        details = ComponentDetails(ccId=ccId, formula=formula, ifCharge=ifCharge)
        descriptors = ComponentDescriptors(smiles=smiles, isoSmiles=isoSmiles, inchi=inchi, inchiKey=inchiKey)
        typeCounts = defaultdict(int)
        ccAtomD = {}
        ccAtomIdD = {}
        ccNameTypeD = {}
        for atId, at in enumerate(self.__oeMol.GetAtoms(), 1):
            atName = at.GetName()
            atNo = at.GetAtomicNum()
            aType = PdbxChemCompConstants.periodicTable[atNo - 1]
            ccNameTypeD[atName] = aType
            if filterHydrogens and aType == "H":
                continue
            typeCounts[aType] += 1
            isAromatic = at.IsAromatic()
            isChiral = at.IsChiral()
            iCharge = at.GetFormalCharge()
            #
            cipStereo = None
            if at.HasStereoSpecified():
                cip = oechem.OEPerceiveCIPStereo(self.__oeMol, at)
                if cip == oechem.OECIPAtomStereo_S:
                    cipStereo = "S"
                if cip == oechem.OECIPAtomStereo_R:
                    cipStereo = "R"
                if cip == oechem.OECIPAtomStereo_NotStereo:
                    cipStereo = "?"
                if isChiral and cip == oechem.OECIPAtomStereo_UnspecStereo:
                    cipStereo = "?"

                #
                if cipStereo and cipStereo not in ["S", "R"]:
                    logger.error("%s (%s): Unexpected atom CIP stereo setting %r", ccId, atName, cipStereo)
            #
            ccAtomD[atName] = ComponentAtom(name=atName, aType=aType, isAromatic=isAromatic, isChiral=isChiral, CIP=cipStereo, fCharge=iCharge)
            ccAtomIdD[atId] = atName
        #
        ccBondD = {}
        for bnd in self.__oeMol.GetBonds():
            atNameI = bnd.GetBgn().GetName()
            atNameJ = bnd.GetEnd().GetName()
            if filterHydrogens and (ccNameTypeD[atNameI] == "H" or ccNameTypeD[atNameJ] == "H"):
                continue
            isAromatic = bnd.IsAromatic()
            iType = bnd.GetOrder()
            cipStereo = None
            if iType == 2:
                cip = oechem.OEPerceiveCIPStereo(self.__oeMol, bnd)
                if bnd.HasStereoSpecified():
                    if cip == oechem.OECIPBondStereo_E:
                        cipStereo = "E"
                    if cip == oechem.OECIPBondStereo_Z:
                        cipStereo = "Z"
                    if cip == oechem.OECIPBondStereo_NotStereo:
                        cipStereo = "?"
                    if cip == oechem.OECIPBondStereo_UnspecStereo:
                        cipStereo = "?"

            if cipStereo and cipStereo not in ["E", "Z"]:
                logger.error("%s (%s %s): Unexpected bond CIP stereo setting %r", ccId, atNameI, atNameJ, cipStereo)
            #
            ccBondD[(atNameI, atNameJ)] = ComponentBond(iType=iType, isAromatic=isAromatic, CIP=cipStereo)
        #
        ccD = {"details": details, "descriptors": descriptors, "atoms": ccAtomD, "bonds": ccBondD}
        return ccD

    # ----
    def buildRelated(self, limitPerceptions=False):
        """ Build the most "authorative" molecule from the chemical component definition and
        use this as a basis reference descriptors and related protomeric and tautomeric forms.
        This collection is designed to capture the chemical diversity within the component defintion
        as well as other reasonable chemical forms for search purposes.  The collection is returned
        as a nested dictionary of conventionally identified SMILES descriptors.

        Returns:
            dict: {'ccId|qualifier': {'smiles': ..., 'inchi-key': ...}}

            qualifier is provided if build_type is not model-xyz

            qualifier m = hashlib.sha256(isoSmiles).hexdigest()
        """
        oeVersionString = "OpenEye OEChem %s" % oechem.OEToolkitsGetRelease()
        retD = {}
        if not self.__isDefinitionSet:
            return retD

        uniqSmilesD = {}
        try:
            for buildType in ["model-xyz", "oe-iso-smiles", "cactvs-iso-smiles", "inchi"]:
                ok = self.build(molBuildType=buildType, setTitle=True, limitPerceptions=limitPerceptions)
                if ok:
                    # name = self.__ccId + "|" + "ref" if buildType == "model-xyz" else self.__ccId + "|" + buildType
                    inchiKey = self.getInChIKey()
                    smiles = self.getIsoSMILES()
                    qualifier = hashlib.sha256(smiles.encode("utf-8")).hexdigest()
                    name = self.__ccId if buildType == "model-xyz" else self.__ccId + "|" + qualifier
                    formula = self.getFormula()
                    fCharge = self.getFormalCharge()
                    eleD = self.getElementCounts(addExplicitHydrogens=True, useSymbol=True)
                    if smiles and inchiKey and smiles not in uniqSmilesD:
                        uniqSmilesD[smiles] = True
                        retD[name] = {
                            "name": name,
                            "build-type": buildType,
                            "smiles": smiles,
                            "inchi-key": inchiKey,
                            "formula": formula,
                            "fcharge": fCharge,
                            "type_counts": eleD,
                            "program": oeVersionString,
                        }
            for buildType in ["oe-smiles", "acdlabs-smiles", "cactvs-smiles"]:
                ok = self.build(molBuildType=buildType, setTitle=True, limitPerceptions=limitPerceptions)
                if ok:
                    inchiKey = self.getInChIKey()
                    smiles = self.getCanSMILES()
                    qualifier = hashlib.sha256(smiles.encode("utf-8")).hexdigest()
                    name = self.__ccId + "|" + qualifier
                    formula = self.getFormula()
                    fCharge = self.getFormalCharge()
                    eleD = self.getElementCounts(addExplicitHydrogens=True, useSymbol=True)
                    if smiles and inchiKey and smiles not in uniqSmilesD:
                        uniqSmilesD[smiles] = True
                        retD[name] = {
                            "name": name,
                            "build-type": buildType,
                            "smiles": smiles,
                            "inchi-key": inchiKey,
                            "formula": formula,
                            "fcharge": fCharge,
                            "type_counts": eleD,
                            "program": oeVersionString,
                        }
            # --- do charge and tautomer normalization on the model-xyz build
            ok = self.build(molBuildType="model-xyz", setTitle=True, limitPerceptions=limitPerceptions)
            if ok:
                logger.debug("%s begin protomer search", self.__ccId)
                upMol = self.getUniqueProtomerMolExtended(maxTautomerAtoms=100, maxSearchTime=0.50)
                if not upMol:
                    logger.warning("%s protomer and tautomer generation failed", self.__ccId)
                else:
                    self.__oeMol = upMol
                    inchiKey = self.getInChIKey()
                    smiles = self.getIsoSMILES()
                    qualifier = hashlib.sha256(smiles.encode("utf-8")).hexdigest()
                    name = self.__ccId + "|" + qualifier
                    formula = self.getFormula()
                    fCharge = self.getFormalCharge()
                    eleD = self.getElementCounts(addExplicitHydrogens=True, useSymbol=True)
                    if smiles and inchiKey and smiles not in uniqSmilesD:
                        uniqSmilesD[smiles] = True
                        retD[name] = {
                            "name": name,
                            "build-type": "unique-protomer|model-xyz",
                            "smiles": smiles,
                            "inchi-key": inchiKey,
                            "formula": formula,
                            "fcharge": fCharge,
                            "type_counts": eleD,
                            "program": oeVersionString,
                        }
                    logger.debug("%s begin tautomer search", self.__ccId)
                    tautomerList = self.getTautomerMolList()
                    logger.debug("%s tautomer count %d", self.__ccId, len(tautomerList))
                    for ii, tMol in enumerate(tautomerList, 1):
                        if tMol:
                            self.__oeMol = tMol
                            smiles = self.getIsoSMILES()
                            inchiKey = self.getInChIKey()
                            qualifier = hashlib.sha256(smiles.encode("utf-8")).hexdigest()
                            label = "tautomer_%d|model-xyz" % ii
                            name = self.__ccId + "|" + qualifier

                            formula = self.getFormula()
                            fCharge = self.getFormalCharge()
                            eleD = self.getElementCounts(addExplicitHydrogens=True, useSymbol=True)
                            if smiles and inchiKey and smiles not in uniqSmilesD:
                                uniqSmilesD[smiles] = True
                                retD[name] = {
                                    "name": name,
                                    "build-type": label,
                                    "smiles": smiles,
                                    "inchi-key": inchiKey,
                                    "formula": formula,
                                    "fcharge": fCharge,
                                    "type-counts": eleD,
                                    "program": oeVersionString,
                                }

        except Exception as e:
            logger.info("Failing for %r with %s", self.__ccId, str(e))

        return retD

    # ----
    def build(self, molBuildType="model", setTitle=True, limitPerceptions=False, fallBackBuildType="model-xyz", normalize=False):
        try:
            if molBuildType in ["ideal-xyz", "model-xyz"]:
                oeMol = self.__build3D(molBuildType=molBuildType, setTitle=setTitle)
            elif molBuildType in ["connection-table"]:
                oeMol = self.__build2D(setTitle=setTitle)
            elif molBuildType in ["oe-iso-smiles", "oe-smiles", "acdlabs-smiles", "cactvs-iso-smiles", "cactvs-smiles", "inchi"]:
                oeMol = self.__buildFromDescriptor(self.__ccId, molBuildType, limitPerceptions=limitPerceptions, fallBackBuildType=fallBackBuildType)
            #
            if normalize:
                nMol = self.getUniqueProtomerMol(oeMol)
                self.__oeMol = nMol if nMol else oeMol
            else:
                self.__oeMol = oeMol
            return oeMol is not None
        except Exception as e:
            logger.info("Failing with %s", str(e))
        return False

    def __buildFromDescriptor(self, ccId, molBuildType, setTitle=True, limitPerceptions=False, fallBackBuildType="model-xyz", rebuildOnFailure=False):
        """Parse the input descriptor string and return a molecule object (OeGraphMol).

        Args:
            ccId (str): chemical component identifier
            molBuildType (str):  source data used to build molecule
            setTitle (bool): make chemical component identifer the molecular title
            limitPerceptions (bool): flag to limit the perceptions/transformations of input SMILES
            fallBackBuildType (str): fallback build type for missing descriptors  (Default:"model-xyz")

        Returns:
            object: OeGraphMol() object or None for failure

        """
        try:
            if molBuildType not in ["oe-iso-smiles", "oe-smiles", "acdlabs-smiles", "cactvs-iso-smiles", "cactvs-smiles", "inchi"]:
                logger.error("%s unexpected molBuildType %r", ccId, molBuildType)
                return None
            #
            ccId = self.__ccId
            #
            descr = self.__getDescriptor(ccId, molBuildType, fallBackBuildType=fallBackBuildType)
            if not descr:
                logger.warning("%r molBuildType %r missing descriptor", ccId, molBuildType)
                return None
            #
            oeMol = oechem.OEGraphMol()
            ok = True
            if molBuildType in ["oe-smiles", "oe-iso-smiles", "acdlabs-smiles", "cactvs-smiles", "cactvs-iso-smiles"]:
                smiles = descr
                if limitPerceptions:
                    # convert the descriptor string into a molecule
                    if not oechem.OEParseSmiles(oeMol, smiles, False, False):
                        logger.warning("%r parsing input failed for %r string", ccId, molBuildType)
                        ok = False
                        if self.__isDefinitionSet and rebuildOnFailure:
                            # Try again with a descriptor rebuild
                            smiles = self.__rebuildDescriptor(ccId, molBuildType, fallBackBuildType=fallBackBuildType)
                            ok = oechem.OEParseSmiles(oeMol, smiles, False, False)
                            if not ok:
                                logger.warning("%s regenerating %r failed from fallback build %r", ccId, molBuildType, fallBackBuildType)
                else:
                    logger.debug("Building with %s", smiles)
                    if not oechem.OESmilesToMol(oeMol, smiles):
                        logger.warning("%r converting input failed for %r string", ccId, molBuildType)
                        ok = False
                        if self.__isDefinitionSet and rebuildOnFailure:
                            # Try again with a descriptor rebuild
                            smiles = self.__rebuildDescriptor(ccId, molBuildType, fallBackBuildType=fallBackBuildType)
                            ok = oechem.OESmilesToMol(oeMol, smiles)
                            if not ok:
                                logger.warning("%s regenerating %r failed from fallback build %r", ccId, molBuildType, fallBackBuildType)

            elif molBuildType in ["inchi"]:
                inchi = descr
                logger.debug("Building with molBuildType %r descr %r ", molBuildType, descr)
                if limitPerceptions:
                    # convert the InChI string into a molecule
                    if not oechem.OEParseInChI(oeMol, inchi):
                        logger.warning("%r parsing input failed for InChI string", ccId)
                        ok = False
                        if self.__isDefinitionSet and rebuildOnFailure:
                            # Try again with a descriptor rebuild
                            inchi = self.__rebuildDescriptor(ccId, molBuildType, fallBackBuildType=fallBackBuildType)
                            ok = oechem.OEParseInChI(oeMol, inchi)
                            if not ok:
                                logger.warning("%s regenerating %r failed from fallback build %r", ccId, molBuildType, fallBackBuildType)
                else:
                    if not oechem.OEInChIToMol(oeMol, inchi):
                        logger.warning("%r converting input failed for InChI string", ccId)
                        ok = False
                        if self.__isDefinitionSet and rebuildOnFailure:
                            # Try again with a descriptor rebuild
                            inchi = self.__rebuildDescriptor(ccId, molBuildType, fallBackBuildType=fallBackBuildType)
                            ok = oechem.OEInChIToMol(oeMol, inchi)
                            if not ok:
                                logger.warning("%s regenerating %r failed from fallback build %r", ccId, molBuildType, fallBackBuildType)
            okT = True
            if ok:
                if setTitle:
                    oeMol.SetTitle(self.__ccId)
                oechem.OETriposAtomNames(oeMol)
                okT = self.__testDescriptorBuild(oeMol)
            #
            return oeMol if ok and okT else None
        except Exception as e:
            logger.exception("%r failing with %s", self.__ccId, str(e))
        return None

    def __testDescriptorBuild(self, oeMol):
        ok = False
        try:
            ok = oeMol.GetTitle() and oechem.OEMolecularFormula(oeMol) and oechem.OECreateIsoSmiString(oeMol)
            logger.debug("%s     Title %r", self.__ccId, oeMol.GetTitle())
            logger.debug("%s   Formula %r", self.__ccId, oechem.OEMolecularFormula(oeMol))
            logger.debug("%s ISOSMILES %r", self.__ccId, oechem.OECreateIsoSmiString(oeMol))
        except Exception as e:
            logger.error("%r failing with %s", self.__ccId, str(e))
        return ok

    def __getDescriptor(self, ccId, molBuildType, fallBackBuildType="model-xyz"):
        if molBuildType not in ["oe-iso-smiles", "oe-smiles", "acdlabs-smiles", "cactvs-iso-smiles", "cactvs-smiles", "inchi"]:
            logger.error("%s unexpected molBuildType %r", ccId, molBuildType)
            return None
        ret = None
        if not self.__isDefinitionSet:
            if molBuildType.upper() == self.__descriptorBuildType.upper():
                ret = self.__descriptor
        else:
            desIt = PdbxChemCompDescriptorIt(self.__dataContainer)
            for des in desIt:
                desBuildType = des.getMolBuildType()
                desText = des.getDescriptor()
                if molBuildType.upper() == desBuildType.upper():
                    ret = desText.strip()
            #
            # Try to create a missing descriptor if possible.
            if ret is None and molBuildType in ["oe-iso-smiles", "oe-smiles", "inchi"] and fallBackBuildType:
                try:
                    logger.info("%r missing descriptor for %r rebuilding using %r", ccId, molBuildType, fallBackBuildType)
                    ret = self.__rebuildDescriptor(ccId, molBuildType, fallBackBuildType=fallBackBuildType)
                except Exception as e:
                    logger.exception("%r descriptor rebuild failed using %r with %s", ccId, fallBackBuildType, str(e))

        return ret

    def __rebuildDescriptor(self, ccId, molBuildType, fallBackBuildType="ideal-xyz"):
        ok = self.build(molBuildType=fallBackBuildType)
        #
        descr = None
        if ok and molBuildType in ["oe-iso-smiles"]:
            descr = self.getIsoSMILES()
        elif ok and molBuildType in ["oe-smiles"]:
            descr = self.getCanSMILES()
        elif ok and molBuildType in ["inchi"]:
            descr = self.getInChI()
        else:
            logger.error("Rebuild failing for %r %r", ccId, molBuildType)
        #
        ret = descr.strip() if descr else None
        if not ret:
            logger.info("%s regenerating descriptor failed for buildType %r", ccId, fallBackBuildType)
        return ret

    def __build2D(self, setTitle=True):
        """  Build molecule using only atom and bond types, and associated annotations
             in a chemical component definition.
        """
        try:
            self.__oeClear()
            oeMol = oechem.OEGraphMol()
            if setTitle:
                oeMol.SetTitle(self.__ccId)
            aL = []
            i = 1
            # Atom index dictionary
            aD = {}
            eD = {}

            atomIt = PdbxChemCompAtomIt(self.__dataContainer)
            for ccAt in atomIt:
                atName = ccAt.getName()
                aD[atName] = i
                i += 1
                atNo = ccAt.getAtNo()
                if atNo not in eD:
                    eD[atNo] = 1
                else:
                    eD[atNo] += 1
                atType = ccAt.getType()
                fc = ccAt.getFormalCharge()
                chFlag = ccAt.isChiral()
                arFlag = ccAt.isAromatic()
                isotope = ccAt.getIsotope()
                leavingAtom = ccAt.getLeavingAtomFlag()

                oeAt = oeMol.NewAtom(atNo)
                oeAt.SetName(atName)
                oeAt.SetFormalCharge(fc)
                oeAt.SetStringData("pdbx_leaving_atom_flag", leavingAtom)
                oeAt.SetChiral(chFlag)
                oeAt.SetIsotope(isotope)
                oeAt.SetAromatic(arFlag)
                if chFlag:
                    st = ccAt.getCIPStereo()
                    if st == "S" or st == "R":
                        oeAt.SetStringData("StereoInfo", st)
                logger.debug("Atom - %s type %s atno %d isotope %d fc %d chFlag %r", atName, atType, atNo, isotope, fc, chFlag)
                aL.append(oeAt)

            bondIt = PdbxChemCompBondIt(self.__dataContainer)
            for ccBnd in bondIt:
                (at1, at2) = ccBnd.getBond()
                iat1 = aD[at1] - 1
                iat2 = aD[at2] - 1
                iType = ccBnd.getIntegerType()
                arFlag = ccBnd.isAromatic()
                logger.debug(" %s %d -- %s %d (%d)", at1, iat1, at2, iat2, iType)

                oeBnd = oeMol.NewBond(aL[iat1], aL[iat2], iType)
                oeBnd.SetAromatic(arFlag)
                if arFlag:
                    oeBnd.SetIntType(5)
                st = ccBnd.getStereo()
                if st == "E" or st == "Z":
                    oeBnd.SetStringData("StereoInfo", st)

            #
            oeMol = self.updateOePerceptions2D(oeMol, aromaticModel=None)
            # run standard perceptions --
            # oechem.OEFindRingAtomsAndBonds(oeMol)
            # oechem.OEPerceiveChiral(oeMol)

            for oeAt in oeMol.GetAtoms():
                st = oeAt.GetStringData("StereoInfo")
                if st == "R":
                    oechem.OESetCIPStereo(oeMol, oeAt, oechem.OECIPAtomStereo_R)
                elif st == "S":
                    oechem.OESetCIPStereo(oeMol, oeAt, oechem.OECIPAtomStereo_S)

            for oeBnd in oeMol.GetBonds():
                st = oeBnd.GetStringData("StereoInfo")
                if st == "E":
                    oechem.OESetCIPStereo(oeMol, oeBnd, oechem.OECIPBondStereo_E)
                elif st == "Z":
                    oechem.OESetCIPStereo(oeMol, oeBnd, oechem.OECIPBondStereo_Z)
            if self.__verbose:
                for ii, atm in enumerate(oeMol.GetAtoms()):
                    logger.debug("OeBuildMol.build2d - atom  %d %s", ii, atm.GetName())
            # JDW Add new
            # self.__transferAromaticFlagsOE(oeMol)
            return oeMol
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return None

    def __build3D(self, molBuildType, setTitle=True, useFallBackModelXyz=True):
        """ Build molecule using only atom and bond types, and aromatic annotations
            in a chemical component defintion. Use the indicated 3D coordinate data
            to assign stereochemistry assignments.
        """
        try:
            self.__oeClear()
            oeMol = oechem.OEGraphMol()
            # oeMol = oechem.OEMol()
            #
            if setTitle:
                oeMol.SetTitle(self.__ccId)
            aL = []

            # Atom index dictionary
            aD = {}
            eD = {}
            i = 1

            atomIt = PdbxChemCompAtomIt(self.__dataContainer)
            for ccAt in atomIt:

                atName = ccAt.getName()
                aD[atName] = i
                i += 1
                atNo = ccAt.getAtNo()
                if atNo not in eD:
                    eD[atNo] = 1
                else:
                    eD[atNo] += 1

                atType = ccAt.getType()
                fc = ccAt.getFormalCharge()
                chFlag = ccAt.isChiral()
                arFlag = ccAt.isAromatic()
                isotope = ccAt.getIsotope()
                leavingAtom = ccAt.getLeavingAtomFlag()

                oeAt = oeMol.NewAtom(atNo)
                oeAt.SetName(atName)
                oeAt.SetFormalCharge(fc)
                oeAt.SetStringData("pdbx_leaving_atom_flag", leavingAtom)
                oeAt.SetChiral(chFlag)
                oeAt.SetIsotope(isotope)
                oeAt.SetAromatic(arFlag)
                tChk = oeAt.IsAromatic()
                if tChk != arFlag:
                    logger.info("%s inconsistent aromatic setting on %s", self.__ccId, atName)

                cTup = None
                if (molBuildType == "model-xyz") and ccAt.hasModelCoordinates():
                    cTup = ccAt.getModelCoordinates()
                    oeMol.SetCoords(oeAt, cTup)
                elif (molBuildType == "ideal-xyz") and ccAt.hasIdealCoordinates():
                    cTup = ccAt.getIdealCoordinates()
                    oeMol.SetCoords(oeAt, cTup)
                elif (molBuildType == "ideal-xyz") and ccAt.hasModelCoordinates() and useFallBackModelXyz:
                    cTup = ccAt.getModelCoordinates()
                    oeMol.SetCoords(oeAt, cTup)
                else:
                    pass

                logger.debug("Atom - %s type %s atno %d isotope %d fc %d (xyz) %r", atName, atType, atNo, isotope, fc, cTup)
                aL.append(oeAt)

            bondIt = PdbxChemCompBondIt(self.__dataContainer)
            for ccBnd in bondIt:
                (at1, at2) = ccBnd.getBond()
                iat1 = aD[at1] - 1
                iat2 = aD[at2] - 1
                iType = ccBnd.getIntegerType()
                logger.debug(" %s %d -- %s %d (%d)", at1, iat1, at2, iat2, iType)
                oeBnd = oeMol.NewBond(aL[iat1], aL[iat2], iType)
                atName1 = oeBnd.GetBgn().GetName()
                atName2 = oeBnd.GetEnd().GetName()
                logger.debug("%s - created bond %s %s", self.__ccId, atName1, atName2)

            #
            # run standard OE perceptions --
            #
            self.updateOePerceptions3D(oeMol)
            # Reset aromatic flags
            # self.__transferAromaticFlagsOE(oeMol)
            return oeMol
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return None

    #
    def getTautomerMolList(self, oeMol=None, maxTautomerAtoms=200, maxSearchTime=60):
        tautomerMolL = []
        tautomerOptions = oequacpac.OETautomerOptions()
        tautomerOptions.SetMaxTautomericAtoms(maxTautomerAtoms)
        tautomerOptions.SetMaxSearchTime(maxSearchTime)
        logger.debug("Tautomer option max atoms = %r", tautomerOptions.GetMaxTautomericAtoms())
        pKaNorm = True
        inMol = oeMol if oeMol else self.__oeMol
        #
        errfs = oechem.oeosstream()
        oechem.OEThrow.SetLevel(oechem.OEErrorLevel_Info)
        oechem.OEThrow.SetOutputStream(errfs)
        oechem.OEThrow.Clear()
        #
        for tautomer in oequacpac.OEGetReasonableTautomers(inMol, tautomerOptions, pKaNorm):
            if "Warning:" in str(errfs.str()):
                logger.info("%s caught OE warning - skipping - %r", self.__ccId, errfs.str()[:-1])
            else:
                tautomerMolL.append(tautomer)
            oechem.OEThrow.Clear()
        # Restore stream error state
        oechem.OEThrow.SetOutputStream(oechem.oeerr)
        oechem.OEThrow.SetLevel(self.__oeErrorLevel)
        #
        return tautomerMolL

    def getUniqueProtomerMol(self, oeMol=None):
        inMol = oeMol if oeMol else self.__oeMol
        mol = oechem.OEGraphMol()
        ok = oequacpac.OEGetUniqueProtomer(mol, inMol)
        return mol if ok else None

    def getNeutralProtomerMol(self, oeMol=None):
        inMol = oeMol if oeMol else oechem.OEGraphMol(self.__oeMol)
        ok = oequacpac.OESetNeutralpHModel(inMol)
        return inMol if ok else None

    def getUniqueProtomerMolExtended(self, oeMol=None, maxTautomerAtoms=200, maxSearchTime=60):
        try:
            inMol = oeMol if oeMol else oechem.OEGraphMol(self.__oeMol)
            oequacpac.OEHypervalentNormalization(inMol)
            oequacpac.OERemoveFormalCharge(inMol)
            oechem.OESuppressHydrogens(inMol)
            opts = oequacpac.OETautomerOptions()
            #
            maxZone = 31
            opts.SetMaxZoneSize(maxZone)
            maxAtoms = maxTautomerAtoms
            opts.SetMaxTautomericAtoms(maxAtoms)
            opts.SetRankTautomers(False)
            opts.SetMaxTautomersGenerated(8192)
            opts.SetMaxTautomersToReturn(1)
            # opts.SetMaxSearchTime(0) #can't use time in canonical process
            opts.SetMaxSearchTime(maxSearchTime)
            opts.SetSaveStereo(False)
            opts.SetRacemicType(oequacpac.OERacemicType_EverSampled)
            #
            errfs = oechem.oeosstream()
            oechem.OEThrow.SetLevel(oechem.OEErrorLevel_Info)
            oechem.OEThrow.SetOutputStream(errfs)
            oechem.OEThrow.Clear()
            #
            tautomerMolL = []
            for tautomer in oequacpac.OEEnumerateTautomers(inMol, opts):
                if "Warning:" in str(errfs.str()):
                    logger.info("%s caught OE warning - skipping - %r", self.__ccId, errfs.str()[:-1])
                else:
                    tautomerMolL.append(tautomer)
                oechem.OEThrow.Clear()
            # Restore stream error state
            oechem.OEThrow.SetOutputStream(oechem.oeerr)
            oechem.OEThrow.SetLevel(self.__oeErrorLevel)
        except Exception as e:
            logger.exception("Failing %r with %s", self.__ccId, str(e))
        #
        outMol = tautomerMolL[0] if tautomerMolL else None
        return outMol if outMol and outMol != inMol else None



##
# File:    MoleculeAnnotationsCompare.py
# Author:  jdw
# Date:    31-Dec-2019
# Version: 0.001
#
# Updates:
#
##
"""
Utilities to compare molecular feature annotations.
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
from collections import defaultdict, namedtuple

from rcsb.utils.chem.PdbxChemComp import PdbxChemCompAtomIt, PdbxChemCompBondIt, PdbxChemCompDescriptorIt, PdbxChemCompIt

# from rcsb.utils.chem.PdbxChemCompConstants import PdbxChemCompConstants


logger = logging.getLogger(__name__)

ComponentDetails = namedtuple("ComponentDetails", "ccId formula ifCharge")
ComponentAtom = namedtuple("ComponentAtom", "name aType isAromatic isChiral CIP fCharge")
ComponentBond = namedtuple("ComponentBond", "iType isAromatic CIP")
ComponentDescriptors = namedtuple("ComponentDescriptors", "smiles isoSmiles inchi inchiKey")
ComponentCompare = namedtuple(
    "ComponentCompare", "ccId refType refInfo tstType tstInfo difDetails difAromaticAtoms difStereoAtoms difTypeBonds difAromaticBonds difSmiles difIsoSmiles difInchi difInchiKey",
)


class MoleculeAnnotationsCompare(object):
    """ Utilities to compare molecular feature annotations.
    """

    def __init__(self, **kwargs):
        _ = kwargs

    def getChemCompFeatures(self, dataContainer, descriptorProgram="OPENEYE", filterHydrogens=False):
        """Get the essential features of the input chemical component definition.
        """
        ccIt = iter(PdbxChemCompIt(dataContainer))
        cc = next(ccIt, None)
        formula = cc.getFormulaWithCharge()
        ccId = cc.getId()
        ccName = cc.getName()
        ifCharge = cc.getFormalChargeAsInt()
        isAmbiguous = cc.getAmbiguousFlag() in ["Y", "y"]
        isCurrent = cc.getReleaseStatus() in ["REL"]
        #
        desIt = PdbxChemCompDescriptorIt(dataContainer)
        isoSmiles = smiles = inchi = inchiKey = None
        for des in desIt:
            desType = des.getType().upper()
            desProg = des.getProgram().upper()
            desText = des.getDescriptor().strip()
            if "OPEN" in desProg and desType == "SMILES_CANONICAL" and descriptorProgram == "OPENEYE":
                isoSmiles = desText
            elif "OPEN" in desProg and desType == "SMILES" and descriptorProgram == "OPENEYE":
                smiles = desText
            elif "CAC" in desProg and desType == "SMILES_CANONICAL" and descriptorProgram == "CACTVS":
                isoSmiles = desText
            elif "CAC" in desProg and desType == "SMILES" and descriptorProgram == "CACTVS":
                smiles = desText
            elif desType == "INCHI":
                inchi = desText
            elif desType == "INCHIKEY":
                inchiKey = desText
            logger.debug("type %r prog %r text %r", desType, desProg, desText)
        logger.debug("smiles %r isosmiles %r inchi %r inchikey %r", smiles, isoSmiles, inchi, inchiKey)
        #

        descriptors = ComponentDescriptors(smiles=smiles, isoSmiles=isoSmiles, inchi=inchi, inchiKey=inchiKey)
        #
        atIt = PdbxChemCompAtomIt(dataContainer)
        typeCounts = defaultdict(int)
        ccNameTypeD = {}
        ccAtomD = {}
        sumCharge = 0
        for at in atIt:
            atName = at.getName()
            aType = at.getType().upper()
            ccNameTypeD[atName] = aType
            if filterHydrogens and aType == "H":
                continue
            #
            ccNameTypeD[atName] = aType
            typeCounts[aType] += 1
            isAromatic = at.isAromatic()
            isChiral = at.isChiral()
            cipStereo = at.getCIPStereo()
            #
            iCharge = at.getFormalChargeAsInt()
            sumCharge += iCharge
            ccAtomD[atName] = ComponentAtom(name=atName, aType=aType, isAromatic=isAromatic, isChiral=isChiral, CIP=cipStereo, fCharge=iCharge)
        #
        if sumCharge != ifCharge and ifCharge == 0:
            ifCharge = sumCharge
            formula = cc.getFormulaWithCharge(ifCharge=ifCharge)

        details = ComponentDetails(ccId=ccId, formula=formula, ifCharge=ifCharge)
        ccBondD = {}
        bndIt = PdbxChemCompBondIt(dataContainer)
        for bnd in bndIt:
            atIdI, atIdJ = bnd.getBond()
            if filterHydrogens and (ccNameTypeD[atIdI] == "H" or ccNameTypeD[atIdJ] == "H"):
                continue
            cipStereo = bnd.getStereo()
            isAromatic = bnd.isAromatic()
            iType = bnd.getIntegerType()
            ccBondD[(atIdI, atIdJ)] = ComponentBond(iType=iType, isAromatic=isAromatic, CIP=cipStereo)
        #
        ccD = {"name": ccName, "isCurrent": isCurrent, "isAmbiguous": isAmbiguous, "details": details, "descriptors": descriptors, "atoms": ccAtomD, "bonds": ccBondD}
        return ccD

    def compare(self, refFD, tstFD, refType="CC", refInfo=None, tstType="CACTVS", tstInfo=None):
        """Compare reference and test molecular features in the input feature dictionaries.

        Args:
            refFD (dict): reference molecule feature dictionary
            tstFD (dict): test molecule feature dictionary
            refType (str, optional): reference molecule type. Defaults to "CC".
            tstType (str, optional): test molecule type. Defaults to "CACTVS".
            refInfo (str, optional): ref molecule additional details. Defaults to None.
            tstInfo (str, optional): test molecule additional details. Defaults to None.


        Returns:
            bool, tuple: flag for no differences, tuple describing comparison details (ComponentCompare)
        """
        compareEachAtom = compareEachBond = False
        difDetails = difAromaticAtoms = difStereoAtoms = difSmiles = difIsoSmiles = difInchi = difInchiKey = False
        difAromaticBonds = difTypeBonds = False
        okR = ok = True
        #
        refDetails = refFD["details"]
        ccId = refDetails.ccId
        #
        isAmbiguous = refFD["isAmbiguous"] if "isAmbiguous" in refFD else False
        isCurrent = refFD["isCurrent"] if "isCurrent" in refFD else False
        refName = refFD["name"] if "name" in refFD else ccId
        #

        #
        tstDetails = tstFD["details"]
        if refDetails != tstDetails:
            logger.error("%s: details differ", ccId)
            logger.error("%s REF: %r", ccId, refDetails)
            logger.error("%s TST: %r", ccId, tstDetails)
            ok = False
        #
        refAtomD = refFD["atoms"]
        refTypeCounts = defaultdict(int)
        refAromaticCount = 0
        refChiralCount = 0
        for atName, atTup in refAtomD.items():
            refTypeCounts[atTup.aType] += 1
            if atTup.isChiral:
                refChiralCount += 1
            if atTup.isAromatic:
                refAromaticCount += 1
        #
        tstAtomD = tstFD["atoms"]
        tstTypeCounts = defaultdict(int)

        tstAromaticCount = 0
        tstChiralCount = 0
        for atName, atTup in tstAtomD.items():
            tstTypeCounts[atTup.aType] += 1
            if atTup.isChiral:
                tstChiralCount += 1
            if atTup.isAromatic:
                tstAromaticCount += 1
        #
        if len(refTypeCounts) != len(tstTypeCounts):
            logger.info("ref %r tst %r", refTypeCounts, tstTypeCounts)
            logger.error("%s: atom types length mismatch", ccId)
            ok = False
        #
        if refChiralCount != tstChiralCount:
            logger.error("%s: chiral atom count missmatch ref: %d tst: %d", ccId, refChiralCount, tstChiralCount)
            ok = False
            difStereoAtoms = True
        #
        if refAromaticCount != tstAromaticCount:
            logger.error("%s: aromatic atom count missmatch ref: %d tst: %d", ccId, refAromaticCount, tstAromaticCount)
            ok = False
            difAromaticAtoms = True
        #
        okR = okR and ok

        for aType in refTypeCounts:
            try:
                ok = refTypeCounts[aType] == tstTypeCounts[aType]
            except Exception:
                ok = False
                okR = okR and ok
            if not ok:
                logger.error("%s:  atom type counts differ for %r", ccId, aType)
        #
        # Atom by atom -
        #
        if compareEachAtom:
            for atName in refAtomD:
                try:
                    ok = refAtomD[atName] == tstAtomD[atName]
                except Exception:
                    ok = False
                    okR = okR and ok
                if not ok:
                    if atName in tstAtomD:
                        logger.error("%s: atom features differ %r: \n -- REF: %r \n -- TST: %r", ccId, atName, refAtomD[atName], tstAtomD[atName])
                    else:
                        logger.error("%s: atom features differ for %r missing atom in tstMol", ccId, atName)
        # ----
        #
        refBondD = refFD["bonds"]
        refBondTypeCounts = defaultdict(int)
        refBondAromaticCount = 0
        for _, bTup in refBondD.items():
            refBondTypeCounts[bTup.iType] += 1
            if bTup.isAromatic:
                refBondAromaticCount += 1
        #
        tstBondD = tstFD["bonds"]
        tstBondTypeCounts = defaultdict(int)
        tstBondAromaticCount = 0
        for _, bTup in tstBondD.items():
            tstBondTypeCounts[bTup.iType] += 1
            if bTup.isAromatic:
                tstBondAromaticCount += 1
        #
        for bType in refBondTypeCounts:
            if bType not in tstBondTypeCounts or refBondTypeCounts[bType] != tstBondTypeCounts[bType]:
                difTypeBonds = True
        #
        if difTypeBonds or len(refBondTypeCounts) != len(tstBondTypeCounts):
            logger.error("%s: bond type mismatch", ccId)
            ok = False
            difTypeBonds = True
        #
        if refBondAromaticCount != tstBondAromaticCount:
            logger.error("%s: aromatic bond count missmatch ref: %d tst: %d", ccId, refBondAromaticCount, tstBondAromaticCount)
            ok = False
            difAromaticBonds = True
        #
        okR = okR and ok
        #
        # Bond by Bond-
        #
        if compareEachBond:
            for ky in refBondD:
                (atNameI, atNameJ) = ky
                try:
                    if (atNameI, atNameJ) in tstBondD:
                        ok = refBondD[(atNameI, atNameJ)] == tstBondD[(atNameI, atNameJ)]
                    elif (atNameJ, atNameI) in tstBondD:
                        ok = refBondD[(atNameI, atNameJ)] == tstBondD[(atNameJ, atNameI)]
                    else:
                        ok = False
                except Exception:
                    ok = False
                #
                okR = okR and ok
                if not ok:
                    if (atNameI, atNameJ) in tstBondD:
                        logger.error(
                            "%s: bond features differ (%r, %r): \n -- REF: %r \n -- TST: %r", ccId, atNameI, atNameJ, refBondD[(atNameI, atNameJ)], tstBondD[(atNameI, atNameJ)]
                        )
                    elif (atNameJ, atNameI) in tstBondD:
                        logger.error(
                            "%s: bond features differ (%r, %r): \n -- REF: %r \n -- TST: %r", ccId, atNameI, atNameJ, refBondD[(atNameI, atNameJ)], tstBondD[(atNameJ, atNameI)]
                        )
                    else:
                        logger.error("%s: bond features differ for (%r, %r) missing bond in tstMol", ccId, atNameI, atNameJ)

        if not okR:
            if not isCurrent or isAmbiguous:
                logger.error("%s: comparison failing - ambiguous flag %r current flag %r name %r", ccId, isAmbiguous, isCurrent, refName)

        #
        # Now do the descriptors
        #
        refDesD = refFD["descriptors"]
        tstDesD = tstFD["descriptors"]
        #
        if refDesD.smiles != tstDesD.smiles:
            logger.error("%s ISOSMILES differ \n -- REF: %r \n -- TST: %r", ccId, refDesD.smiles, tstDesD.smiles)
            difSmiles = True
        if refDesD.isoSmiles != tstDesD.isoSmiles:
            logger.error("%s ISOSMILES differ \n -- REF: %r \n -- TST: %r", ccId, refDesD.isoSmiles, tstDesD.isoSmiles)
            difIsoSmiles = True
        if refDesD.inchi != tstDesD.inchi and tstDesD.inchi:
            logger.error("%s InChIs differ", ccId)
            difInchi = True
        if refDesD.inchiKey != tstDesD.inchiKey and tstDesD.inchiKey:
            logger.error("%s InChI Keys differ", ccId)
            difInchiKey = True
        #
        retCmp = ComponentCompare(
            ccId=ccId,
            refType=refType,
            refInfo=refInfo,
            tstType=tstType,
            tstInfo=tstInfo,
            difDetails=difDetails,
            difAromaticAtoms=difAromaticAtoms,
            difStereoAtoms=difStereoAtoms,
            difTypeBonds=difTypeBonds,
            difAromaticBonds=difAromaticBonds,
            difSmiles=difSmiles,
            difIsoSmiles=difIsoSmiles,
            difInchi=difInchi,
            difInchiKey=difInchiKey,
        )
        return okR, retCmp



##
# File:    IoUtils.py
# Author:  jdw
# Date:    31-Dec-2019
# Version: 0.001
#
# Updates:
#
##
"""
Utilities to manage general component IO and format conversion operations.
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import datetime
import logging

from rcsb.utils.chem.PdbxChemComp import PdbxChemCompAtomIt, PdbxChemCompBondIt, PdbxChemCompIt
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class IoUtils(object):
    """ Utility methods to manage OE specific IO and format conversion operations.
    """

    def __init__(self, **kwargs):
        self.__dirPath = kwargs.get("dirPath", ".")
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        #

    def getComponentDefinitions(self, ccdFilePath):
        try:
            rdCcObjL = self.__mU.doImport(ccdFilePath, fmt="mmcif")
            logger.info("Read %s with %d definitions", ccdFilePath, len(rdCcObjL))
        except Exception as e:
            logger.exception("Loading %s failing with %s", ccdFilePath, str(e))
        return rdCcObjL

    def makeSdf(self, dataContainer, molBuildType="model-xyz", useAromatic=False):
        sdf = ""
        atomIdxD = {}
        ok = True
        try:
            if not (dataContainer.exists("chem_comp") and dataContainer.exists("chem_comp_atom") and dataContainer.exists("chem_comp_bond")):
                return False, sdf, atomIdxD
            ccIt = PdbxChemCompIt(dataContainer)
            for cc in ccIt:
                ccId = cc.getId()
            #
            now = datetime.datetime.now().strftime("%y%m%d%H%M")
            chargeMap = {1: 3, 2: 2, 3: 1, -1: 5, -2: 6, -3: 7, 0: 0}
            # bondMap = {"arom": 4, "delo": 5, "doub": 2, "pi": 5, "poly": 5, "quad": 8, "sing": 1, "trip": 3}
            oL = []
            atomIt = PdbxChemCompAtomIt(dataContainer)
            numAtom = len(atomIt)
            bondIt = PdbxChemCompBondIt(dataContainer)
            numBond = len(bondIt)
            # Header Line 1 - title -- unformatted text --
            oL.append("%s" % ccId)
            # Header Line 2 - xxNNNNNNNNMMDDYYhhmm
            # oL.append("%2s%8s%10s" % ("  ", "CCTOOLS-", now))
            oL.append("%2s%8s%10s" % ("  ", "CHMUTIL-", now))
            # oL.append("  -ISIS-            %s" % "3D")
            #
            oL.append("")
            chiralFlag = 0
            # Connection table - Counts line ---
            sL = "%3d%3d%3d%3d%3d%3d%3d%3d%3d%3d%3d%6s" % (numAtom, numBond, 0, 0, chiralFlag, 0, 0, 0, 0, 0, 999, " V2000")
            oL.append(sL)
            #
            aD = {}
            fcL = []
            isotopeL = []
            iAtom = 1
            atomIt = PdbxChemCompAtomIt(dataContainer)
            for ccAt in atomIt:
                atName = ccAt.getName()
                aD[atName] = iAtom
                atomIdxD[iAtom] = atName
                atNo = ccAt.getAtNo()
                atType = ccAt.getType()
                fc = int(ccAt.getFormalCharge())
                if fc != 0:
                    fcL.append("%4s%4d" % (iAtom, fc))
                # chFlag = ccAt.isChiral()
                # arFlag = ccAt.isAromatic()
                isotope = ccAt.getIsotope()
                if isotope != 0:
                    isotopeL.append("%4s%4d" % (iAtom, isotope))
                cTup = None
                if (molBuildType == "model-xyz") and ccAt.hasModelCoordinates():
                    cTup = ccAt.getModelCoordinates()
                elif (molBuildType == "ideal-xyz") and ccAt.hasIdealCoordinates():
                    cTup = ccAt.getIdealCoordinates()
                else:
                    logger.warning("%s has no %s records", dataContainer.getName(), molBuildType)
                    ok = False
                    cTup = (0.0, 0.0, 0.0)
                iAtom += 1
                #
                aS = "%10.4f%10.4f%10.4f %-3s%2d%3d%3d%3d%3d" % (cTup[0], cTup[1], cTup[2], atType, 0, chargeMap[fc], 0, 0, 0)
                oL.append(aS)
                logger.debug("Atom - %s type %s atno %d isotope %d fc %d (xyz) %r", atName, atType, atNo, isotope, fc, cTup)
                #

            bondIt = PdbxChemCompBondIt(dataContainer)
            for ccBnd in bondIt:
                (at1, at2) = ccBnd.getBond()

                iType = ccBnd.getIntegerType()
                if useAromatic and ccBnd.isAromatic():
                    iType = 4
                bS = "%3d%3d%3d%3d%3d%3d" % (aD[at1], aD[at2], iType, 0, 0, 0)
                oL.append(bS)
                logger.debug(" %s %d -- %s %d (%d)", at1, aD[at1], at2, aD[at2], iType)
            #
            # Add charge line -
            if fcL:
                oL.append("M  CHG%3s%s" % (len(fcL), "".join(fcL)))
            if isotopeL:
                oL.append("M  ISO%3s%s" % (len(isotopeL), "".join(isotopeL)))
            #
            oL.append("M  END")
            oL.append("$$$$")
            #
            sdf = "\n".join(oL)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ok = False

        return ok, sdf, atomIdxD



##
# File:    ChemCompSearchWrapper.py
# Author:  jdw
# Date:    9-Mar-2020
# Version: 0.001
#
# Updates:
#
##
"""
Wrapper for chemical component search operations.
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import copy
import logging
import platform
import resource
import os
import time

from collections import namedtuple

from rcsb.utils.chem.ChemCompIndexProvider import ChemCompIndexProvider
from rcsb.utils.chem.ChemCompSearchIndexProvider import ChemCompSearchIndexProvider
from rcsb.utils.chem.MolecularFormula import MolecularFormula
from rcsb.utils.chem.OeSearchMoleculeProvider import OeSearchMoleculeProvider
from rcsb.utils.chem.OeIoUtils import OeIoUtils
from rcsb.utils.chem.OeSearchUtils import OeSearchUtils
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.SingletonClass import SingletonClass

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logger = logging.getLogger(__name__)

MatchResults = namedtuple("MatchResults", "ccId oeMol searchType matchOpts screenType fpType fpScore oeIdx formula", defaults=(None,) * 9)


class ChemCompSearchWrapper(SingletonClass):
    """ Wrapper for chemical component search operations.
    """

    def __init__(self):
        self.__startTime = time.time()
        self.__cachePath = os.environ["CHEM_SEARCH_CACHE_PATH"]
        self.__ccFileNamePrefix = os.environ["CHEM_SEARCH_CC_PREFIX"]
        # ---
        self.__mU = MarshalUtil(workPath=self.__cachePath)
        # ---
        self.__configD = {}
        self.__ccIdxP = None
        self.__siIdx = {}
        self.__oesmP = None
        self.__oesU = None
        # ---
        self.__statusDescriptorError = -100
        self.__searchError = -200
        self.__searchSuccess = 0

    def readConfig(self):
        #
        ok = False
        try:
            self.__cachePath = os.environ.get("CHEM_SEARCH_CACHE_PATH", ".")
            self.__ccFileNamePrefix = os.environ.get("CHEM_SEARCH_CC_PREFIX", "cc-full")
            configFilePath = os.path.join(self.__cachePath, "config", self.__ccFileNamePrefix + "-config.json")
            configD = self.__mU.doImport(configFilePath, fmt="json")
            logger.debug("ConfigD: %r", configD)
            if configD and (len(configD) > 2) and float(configD["versionNumber"]) > 0.1:
                logger.info("Read version %r sections %r from %s", configD["versionNumber"], list(configD.keys()), configFilePath)
                ok = True
                self.__configD = configD
            else:
                logger.error("Reading config file fails from %r", configFilePath)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ok = False
        return ok

    def updateChemCompIndex(self, useCache=False):
        """Rebuild the basic index of source chemical component and BIRD definitions.
           Update the internal state of this index in the current object instance.

            Resource requirements: 94 sec 1 proc 7GB memory macbook pro

        Args:
            useCache (bool): False to rebuild search index and True to reload

        Returns:
            bool: True for success or false otherwise
        """
        ok = False
        try:
            kwargs = copy.deepcopy(self.__configD["ccsiKwargs"]) if "ccsiKwargs" in self.__configD else None
            if kwargs:
                kwargs["useCache"] = useCache
                ccIdxP = ChemCompIndexProvider(**kwargs)
                ok = ccIdxP.testCache()
                self.__ccIdxP = ccIdxP if ok else None
                logger.info("Chemical component index status %r", ok)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def getChemCompIndex(self):
        return self.__ccIdxP.getIndex() if self.__ccIdxP else {}

    def getSearchMoleculeProvider(self):
        return self.__oesmP if self.__oesmP else None

    def updateSearchIndex(self, useCache=False):
        """Rebuild the search index from source chemical component and BIRD definitions.
           Update the internal state of this index in the current object instance.

            Resource requirements 771 secs 6 proc macbook pro 7GB memory.

        Args:
            useCache (bool): False to rebuild search index and True to reload

        Returns:
            bool: True for success or false otherwise
        """
        ok = False
        try:
            kwargs = copy.deepcopy(self.__configD["ccsiKwargs"]) if "ccsiKwargs" in self.__configD else None
            if kwargs:
                kwargs["useCache"] = useCache
                siIdxP = ChemCompSearchIndexProvider(**kwargs)
                ok = siIdxP.testCache()
                self.__siIdx = siIdxP.getIndex() if siIdxP and ok else {}
                logger.info("Search index status %r index len %d", ok, len(self.__siIdx) if self.__siIdx else 0)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def updateSearchMoleculeProvider(self, useCache=False):
        """Rebuild the search molecule provider.
           Update the internal state of this object reference in the current object instance.

           Resource requirements: 151 seconds 1 proc  0.5GB memory macbook pro

        Args:
            useCache (bool): False to rebuild molecule store and True to reload

        Returns:
            bool: True for success or false otherwise
        """
        ok = False
        try:
            kwargs = copy.deepcopy(self.__configD["oesmpKwargs"]) if "oesmpKwargs" in self.__configD else None
            if kwargs:
                kwargs["useCache"] = useCache
                oesmP = OeSearchMoleculeProvider(**kwargs)
                ok = oesmP.testCache()
                self.__oesmP = oesmP if oesmP and ok else None
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def reloadSearchDatabase(self):
        """Reload the in-memory search databases from the OE molecule provider.
           Resource requirements: ~90sec load time 0.35 GB memory

        Returns:
            bool: True for success or False otherwise
        """
        ok = False
        try:
            okmp = self.updateSearchMoleculeProvider(useCache=True)
            if not okmp:
                return ok
            fpTypeCuttoffD = self.__configD["oesmpKwargs"]["fpTypeCuttoffD"] if "fpTypeCuttoffD" in self.__configD["oesmpKwargs"] else {}
            fpTypeList = [k for k, v in fpTypeCuttoffD.items()]
            oesU = OeSearchUtils(self.__oesmP, fpTypeList=fpTypeList)
            ok = oesU.testCache()
            self.__oesU = oesU if ok else None
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def matchByDescriptor(self, descriptor, descriptorType, matchOpts="graph-relaxed", searchId=None):
        """Return graph match (w/  finger print pre-filtering) and finger print search results for the
           input desriptor.

        Args:
            descriptor (str):  molecular descriptor (SMILES, InChI)
            descriptorType (str): descriptor type (SMILES, InChI
            matchOpts (str, optional): graph match criteria (graph-relaxed, graph-relaxed-stereo, graph-strict, fingerprint-similarity). Defaults to "graph-relaxed".
            searchId (str, optional): search identifier for logging. Defaults to None.

        Returns:
            (statusCode, list, list): status, graph match and finger match lists of type (MatchResults)
                                      -200 descriptor processing error
                                      -100 search execution error
                                         0 search execution success
        """
        ssL = fpL = []
        retStatus = False
        try:
            fpTypeCuttoffD = self.__configD["oesmpKwargs"]["fpTypeCuttoffD"] if "fpTypeCuttoffD" in self.__configD["oesmpKwargs"] else {}
            maxFpResults = self.__configD["oesmpKwargs"]["maxFpResults"] if "maxFpResults" in self.__configD["oesmpKwargs"] else 50
            limitPerceptions = self.__configD["oesmpKwargs"]["limitPerceptions"] if "limitPerceptions" in self.__configD["oesmpKwargs"] else False
            #
            searchId = searchId if searchId else "query"
            messageTag = searchId + ":" + descriptorType
            oeioU = OeIoUtils()
            oeMol = oeioU.descriptorToMol(descriptor, descriptorType, limitPerceptions=limitPerceptions, messageTag=messageTag)
            if not oeMol:
                logger.warning("descriptor type %r molecule build fails: %r", descriptorType, descriptor)
                return self.__statusDescriptorError, ssL, fpL
            #
            retStatus, ssL, fpL = self.__oesU.searchSubStructureAndFingerPrint(oeMol, list(fpTypeCuttoffD.items())[:2], maxFpResults, matchOpts=matchOpts)
            statusCode = 0 if retStatus else self.__searchError
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            #
        return statusCode, ssL, fpL

    def matchByFormulaRange(self, elementRangeD, matchSubset=False, searchId=None):
        """Return formula match results for input element range dictionary.

        Args:
            elementRangeD (dict): {'<element_name>: {'min': <int>, 'max': <int>}, ... }
            matchSubset (bool, optional): query for formula subset (default: False)
            searchId (str, optional): search identifier for logging. Defaults to None.

        Returns:
            (statusCode, list): status, list of chemical component identifiers
        """
        ok = False
        rL = []
        try:
            startTime = time.time()
            searchId = searchId if searchId else "query"
            rL = self.__ccIdxP.matchMolecularFormulaRange(elementRangeD, matchSubset=matchSubset)
            ok = True
            logger.info("%s formula %r matched %d (%.4f seconds)", searchId, elementRangeD, len(rL), time.time() - startTime)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok, rL

    def matchByFormula(self, formula, matchSubset=False, searchId=None):
        """Return formula match results for input molecular formula.

        Args:
            formula (str): molecular formula  (ex. 'C6H6')
            matchSubset (bool, optional): query for formula subset (default: False)
            searchId (str, optional): search identifier for logging. Defaults to None.

        Returns:
            (statusCode, list): status, list of chemical component identifiers
        """
        ok = False
        rL = []
        try:
            startTime = time.time()
            searchId = searchId if searchId else "query"
            mf = MolecularFormula()
            eD = mf.parseFormula(formula)
            elementRangeD = {k.upper(): {"min": v, "max": v} for k, v in eD.items()}
            rL = self.__ccIdxP.matchMolecularFormulaRange(elementRangeD, matchSubset=matchSubset)
            ok = True
            logger.info("%s formula %r matched %d (%.4f seconds)", searchId, elementRangeD, len(rL), time.time() - startTime)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok, rL

    def status(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Status at %s (up %.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)



##
# File:  OeDepict.py
# Date:  25-Oct-2019  J. Westbrook  adapted from earlier code module.
#
# Updates:
#  2-April-2020 jdw enlarge the base image size to avoid dropping labels.
#                   Tweak stereo style to include W/H but avoid undefined stereo marking
#
##
"""
Utilities to depict individual OE molecules.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import os

from openeye import oechem
from openeye import oedepict

logger = logging.getLogger(__name__)


class LabelAtoms(oedepict.OEDisplayAtomPropBase):
    def __init__(self):
        oedepict.OEDisplayAtomPropBase.__init__(self)

    def __call__(self, atom):
        return atom.GetName()

    def CreateCopy(self):
        # __disown__ is required to allow C++ to take
        # ownsership of this object and its memory
        copy = LabelAtoms()
        return copy.__disown__()


class OeDepictBase(object):
    """ Base class for 2D depictions in single and multi-page format.

    """

    def __init__(self):
        super(OeDepictBase, self).__init__()
        self._molTitleList = []
        self._opts = None
        #
        # internal dictionary of display parameters -
        #
        self._params = {
            "imageSizeX": 2500,
            "imageSizeY": 2500,
            "cellBorders": True,
            "suppressHydrogens": False,
            "labelAtomName": False,
            "labelAtomCIPStereo": False,
            "labelAtomIndex": False,
            "labelBondCIPStereo": False,
            "labelBondIndex": False,
            "bondDisplayWidth": None,
            "gridRows": 2,
            "gridCols": 2,
            "cellGap": 5,
            "cellMargin": 10,
            "pageOrientation": "landscape",
            "highlightStyleRef": "none",
            "highlightStyleFit": "ballAndStick",
            "highLightMatchColorRef": "blue",
            "highLightMatchColorFit": "blue",
            "highLightNotMatchColorRef": "pink",
            "highLightNotMatchColorFit": "pink",
        }

    def setDisplayOptions(self, **kwargs):
        self._params.update(kwargs)

    def setGridOptions(self, rows=1, cols=1, cellBorders=True):
        self._params["gridRows"] = rows
        self._params["gridCols"] = cols
        self._params["cellBorders"] = cellBorders

    def setMolTitleList(self, oeMolTitleList):
        """Set the list of OE Mols to be depicted as a list of tuples containing
            [(ccId,oeMol,titleString),(ccId,oeMol,titleString),...]
        """
        self._molTitleList = oeMolTitleList

    def _assignDisplayOptions(self):
        if self._params["labelAtomCIPStereo"]:
            #
            self._opts.SetAtomStereoStyle(oedepict.OEAtomStereoStyle_Display_All)
            # self._opts.SetAtomStereoStyle(oedepict.OEAtomStereoStyle_Display_CIPAtomStereo)

        if self._params["labelBondCIPStereo"]:
            # will include bowties for undefined stereo
            # self._opts.SetBondStereoStyle(oedepict.OEBondStereoStyle_Display_All)
            self._opts.SetBondStereoStyle(oedepict.OEBondStereoStyle_Display_CIPBondStereo)
            self._opts.SetAtomPropLabelFontScale(0.650)

        if self._params["labelAtomIndex"]:
            self._opts.SetAtomPropertyFunctor(oedepict.OEDisplayAtomIdx())
            self._opts.SetAtomPropLabelFont(oedepict.OEFont(oechem.OEDarkGreen))
            self._opts.SetAtomPropLabelFontScale(0.650)

        if self._params["labelBondIndex"]:
            self._opts.SetBondPropertyFunctor(oedepict.OEDisplayBondIdx())
            self._opts.SetBondPropLabelFont(oedepict.OEFont(oechem.OEDarkBlue))

        if self._params["labelAtomName"]:
            atomlabel = LabelAtoms()
            self._opts.SetAtomPropertyFunctor(atomlabel)
            self._opts.SetAtomPropLabelFont(oedepict.OEFont(oechem.OEDarkGreen))
            self._opts.SetAtomPropLabelFontScale(0.650)

        if self._params["bondDisplayWidth"] is not None:
            pen = oedepict.OEPen(oechem.OEBlack, oechem.OEBlack, oedepict.OEFill_On, self._params["bondDisplayWidth"])
            self._opts.SetDefaultBondPen(pen)
            # remove for the moment.  not supported on all platforms
            # self._opts.SetBondWidthScaling(False)
        #
        # 5.0 is minimum size -
        self._opts.SetTitleHeight(5.0)
        #


class OeDepictMultiPage(OeDepictBase):

    """ Create 2D depictions in multipage format from a list of OE molecules and title strings

    """

    def __init__(self, useTitle=True):
        super(OeDepictMultiPage, self).__init__()
        self.__useTitle = useTitle
        self.__multi = None
        self.__image = None
        #

    def __setupImage(self):
        if self._params["pageOrientation"] == "landscape":
            self.__multi = oedepict.OEMultiPageImageFile(oedepict.OEPageOrientation_Landscape, oedepict.OEPageSize_US_Letter)
        else:
            self.__multi = oedepict.OEMultiPageImageFile(oedepict.OEPageOrientation_Portrait, oedepict.OEPageSize_US_Letter)
        self.__image = self.__multi.NewPage()
        self._opts = oedepict.OE2DMolDisplayOptions()

    def prepare(self):
        self.__setupImage()
        rows = self._params["gridRows"]
        cols = self._params["gridCols"]
        grid = oedepict.OEImageGrid(self.__image, rows, cols)

        citer = grid.GetCells()

        for ccId, oeMol, title in self._molTitleList:
            logger.debug("Preparing %s %r", ccId, title)
            if not citer.IsValid():
                # go to next page
                self.__image = self.__multi.NewPage()
                grid = oedepict.OEImageGrid(self.__image, rows, cols)
                grid.SetCellGap(self._params["cellGap"])
                grid.SetMargins(self._params["cellMargin"])
                citer = grid.GetCells()

            cell = citer.Target()
            #
            if self._params["suppressHydrogens"]:
                # mol = oeMol.getGraphMolSuppressH()
                #  OESuppressHydrogens(self.__oeMol, retainPolar=False,retainStereo=True,retainIsotope=True)
                mol = oechem.OESuppressHydrogens(oechem.OEGraphMol(oeMol))
            else:
                mol = oeMol

            if self.__useTitle and title:
                mol.SetTitle(title)
                self._opts.SetTitleHeight(5.0)
            else:
                mol.SetTitle("")
            #
            #
            oedepict.OEPrepareDepiction(mol)
            self._opts.SetDimensions(cell.GetWidth(), cell.GetHeight(), oedepict.OEScale_AutoScale)
            self._assignDisplayOptions()

            disp = oedepict.OE2DMolDisplay(mol, self._opts)
            oedepict.OERenderMolecule(cell, disp)
            oedepict.OEDrawBorder(cell, oedepict.OEPen(oedepict.OEBlackPen))

            citer.Next()

    def write(self, imagePath):
        try:
            dirPath, _ = os.path.split(imagePath)
            if not os.access(dirPath, os.W_OK):
                os.makedirs(dirPath, mode=0o755)
        except Exception:
            pass
        oedepict.OEWriteMultiPageImage(imagePath, self.__multi)


class OeDepict(OeDepictBase):

    """ Create 2D depictions in single-page format from a list of OE molecules & title strings

    """

    def __init__(self, useTitle=True):
        super(OeDepict, self).__init__()
        self.__useTitle = useTitle
        self.__image = None
        self.__grid = None

    def __setupImage(self):
        """ Internal method to configure a single page image.
        """
        #
        self.__image = oedepict.OEImage(self._params["imageSizeX"], self._params["imageSizeY"])
        self.__grid = oedepict.OEImageGrid(self.__image, self._params["gridRows"], self._params["gridCols"])
        self.__grid.SetCellGap(self._params["cellGap"])
        self.__grid.SetMargins(self._params["cellMargin"])
        self._opts = oedepict.OE2DMolDisplayOptions(self.__grid.GetCellWidth(), self.__grid.GetCellHeight(), oedepict.OEScale_AutoScale)
        #
        logger.debug("Num columns %d", self.__grid.NumCols())
        logger.debug("Num rows    %d", self.__grid.NumRows())

    def prepare(self):
        """[summary]
         # OESuppressHydrogens(self.__oeMol, retainPolar=False,retainStereo=True,retainIsotope=True)
        oechem.OESuppressHydrogens(self.__oeMol)
        """
        self.__setupImage()
        for idx, cell in enumerate(self.__grid.GetCells()):
            ccId, oeMol, title = self._molTitleList[idx]
            logger.debug("Preparing %s %r", ccId, title)
            #
            if self._params["suppressHydrogens"]:
                # mol = oeMol.getGraphMolSuppressH()
                #  OESuppressHydrogens(self.__oeMol, retainPolar=False,retainStereo=True,retainIsotope=True)
                mol = oechem.OESuppressHydrogens(oechem.OEGraphMol(oeMol))
            else:
                mol = oeMol
            #
            if self.__useTitle and title:
                mol.SetTitle(title)
                self._opts.SetTitleHeight(5.0)
            else:
                mol.SetTitle("")
            #
            #
            oedepict.OEPrepareDepiction(mol)
            self._opts.SetDimensions(cell.GetWidth(), cell.GetHeight(), oedepict.OEScale_AutoScale)

            self._assignDisplayOptions()

            disp = oedepict.OE2DMolDisplay(mol, self._opts)
            oedepict.OERenderMolecule(cell, disp)
            if self._params["cellBorders"]:
                oedepict.OEDrawBorder(cell, oedepict.OEPen(oedepict.OEBlackPen))

    def write(self, imagePath):
        try:
            dirPath, _ = os.path.split(imagePath)
            if not os.access(dirPath, os.W_OK):
                os.makedirs(dirPath, mode=0o755)
        except Exception:
            pass
        oedepict.OEWriteImage(imagePath, self.__image)



##
# File: PdbxChemComp.py
# Date: 2-Oct-2019  John Westbrook
#
# Update:
#   2-Oct-2019 jdw adapted from PdbxChemCompPersist()
#   7-Nov-2019 jdw add alternate charge and formula methods
##
"""
A collection of access and iterator classes supporting chemical component dictionary data.
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging

from rcsb.utils.chem.PdbxChemCompConstants import PdbxChemCompConstants

logger = logging.getLogger(__name__)


class PdbxCategoryItBase(object):
    """  Base category iterator class.
    """

    def __init__(self, dataCategory, func):
        self.__rL = dataCategory.getRowList() if dataCategory else []
        self.__func = func

    def get(self, index=0):
        try:
            return self.__rL[index]
        except Exception:
            return []

    def __iter__(self):
        return self.forward()

    def __len__(self):
        return len(self.__rL)

    def forward(self):
        # Forward generator
        currentRow = 0
        while currentRow < len(self.__rL):
            row = self.__rL[currentRow]
            currentRow += 1
            yield self.__func(row)

    def reverse(self):
        # The reverse generator
        currentRow = len(self.__rL)
        while currentRow > 0:
            currentRow -= 1
            yield self.__func(self.__rL[currentRow])


class PdbxChemCompIt(PdbxCategoryItBase):
    def __init__(self, dataContainer):
        dataCategory = dataContainer.getObj("chem_comp")
        obj = PdbxChemCompPersist([], attributeNameList=dataCategory.getAttributeList() if dataCategory else [])
        super(PdbxChemCompIt, self).__init__(dataCategory, obj.set)


class PdbxChemCompAtomIt(PdbxCategoryItBase):
    def __init__(self, dataContainer):
        dataCategory = dataContainer.getObj("chem_comp_atom")
        obj = PdbxChemCompAtomPersist([], attributeNameList=dataCategory.getAttributeList() if dataCategory else [])
        super(PdbxChemCompAtomIt, self).__init__(dataCategory, obj.set)


class PdbxChemCompBondIt(PdbxCategoryItBase):
    def __init__(self, dataContainer):
        dataCategory = dataContainer.getObj("chem_comp_bond")
        obj = PdbxChemCompBondPersist([], attributeNameList=dataCategory.getAttributeList() if dataCategory else [])
        super(PdbxChemCompBondIt, self).__init__(dataCategory, obj.set)


class PdbxChemCompDescriptorIt(PdbxCategoryItBase):
    def __init__(self, dataContainer):
        dataCategory = dataContainer.getObj("pdbx_chem_comp_descriptor")
        obj = PdbxChemCompDescriptorPersist([], attributeNameList=dataCategory.getAttributeList() if dataCategory else [])
        super(PdbxChemCompDescriptorIt, self).__init__(dataCategory, obj.set)


class PdbxChemCompIdentifierIt(PdbxCategoryItBase):
    def __init__(self, dataContainer):
        dataCategory = dataContainer.getObj("pdbx_chem_comp_identifier")
        obj = PdbxChemCompIdentifierPersist([], attributeNameList=dataCategory.getAttributeList() if dataCategory else [])
        super(PdbxChemCompIdentifierIt, self).__init__(dataCategory, obj.set)


class PdbxChemCompAuditIt(PdbxCategoryItBase):
    def __init__(self, dataContainer):
        dataCategory = dataContainer.getObj("pdbx_chem_comp_audit")
        obj = PdbxChemCompAuditPersist([], attributeNameList=dataCategory.getAttributeList() if dataCategory else [])
        super(PdbxChemCompAuditIt, self).__init__(dataCategory, obj.set)


class PdbxChemCompPersist(object):
    """ Accessor methods chemical component attributes.

    """

    def __init__(self, rowData, attributeNameList):
        self.__rowData = rowData
        self.__attributeNameList = attributeNameList

    def set(self, rowData=None):
        self.__rowData = rowData
        return self

    def __getAttribute(self, name):
        try:
            i = self.__attributeNameList.index(name)
            return self.__rowData[i]
        except Exception:
            return None

    def getId(self):
        return self.__getAttribute("id")

    def getName(self):
        return self.__getAttribute("name")

    def getType(self):
        return self.__getAttribute("type")

    def getPdbxType(self):
        return self.__getAttribute("pdbx_type")

    def getFormula(self):
        return self.__getAttribute("formula")

    def getSynonyms(self):
        return self.__getAttribute("pdbx_synonyms")

    def getFormalCharge(self):
        return self.__getAttribute("pdbx_formal_charge")

    def getFormalChargeAsInt(self):
        tch = self.getFormalCharge()
        iCharge = int(tch) if tch and tch not in [".", "?"] else 0
        return iCharge

    def getFormulaWithCharge(self, ifCharge=None):
        formula = (self.getFormula()).replace(" ", "")
        fcharge = ifCharge if ifCharge else self.getFormalChargeAsInt()
        if fcharge:
            sign = "+" if fcharge > 0 else "-"
            mag = str(abs(fcharge)) if abs(fcharge) > 1 else ""
            formula = formula + sign + mag
        return formula

    def getModificationDate(self):
        return self.__getAttribute("pdbx_modified_date")

    def getInitialDate(self):
        return self.__getAttribute("pdbx_initial_date")

    def getReleaseStatus(self):
        return self.__getAttribute("pdbx_release_status")

    def getFormulaWeight(self):
        return self.__getAttribute("formula_weight")

    def getSubComponentList(self):
        return self.__getAttribute("pdbx_subcomponent_list")

    def getAmbiguousFlag(self):
        return self.__getAttribute("pdbx_ambiguous_flag")

    def getProcessingSite(self):
        return self.__getAttribute("pdbx_processing_site")

    def getReplacesId(self):
        return self.__getAttribute("pdbx_replaces")

    def getReplacesById(self):
        return self.__getAttribute("pdbx_replaced_by")

    def getNstdParentId(self):
        return self.__getAttribute("mon_nstd_parent_comp_id")

    def getOneLetterCode(self):
        return self.__getAttribute("one_letter_code")

    def getThreeLetterCode(self):
        return self.__getAttribute("three_letter_code")

    def getModelCoordinatesPdbCode(self):
        return self.__getAttribute("pdbx_model_coordinates_db_code")

    def getMissingModelCoordinates(self):
        return self.__getAttribute("pdbx_model_coordinates_missing_flag")

    def getMissingIdealCoordinates(self):
        return self.__getAttribute("pdbx_ideal_coordinates_missing_flag")


class PdbxChemCompAtomPersist(object):
    """ Accessor methods chemical component atom attributes.

    """

    def __init__(self, rowData, attributeNameList):
        self.__rowData = rowData
        self.__attributeNameList = attributeNameList

    def __getAttribute(self, name):
        try:
            i = self.__attributeNameList.index(name)
            return self.__rowData[i]
        except Exception:
            return None

    def set(self, rowData=None):
        self.__rowData = rowData
        return self

    def getName(self):
        return self.__getAttribute("atom_id")

    def isChiral(self):
        return self.__getAttribute("pdbx_stereo_config") != "N"

    def getType(self):
        return self.__getAttribute("type_symbol")

    def getLeavingAtomFlag(self):
        return self.__getAttribute("pdbx_leaving_atom_flag")

    def getAtNo(self):
        try:
            tyU = str(self.getType()).upper()
            if (tyU == "D") or (tyU == "T"):
                tyU = "H"
            return PdbxChemCompConstants.periodicTable.index(tyU) + 1
        except Exception:
            # traceback.print_exc(file=self.__lfh)
            return 0

    def getIsotope(self):
        ty = self.getType()
        if ty == "D":
            return 2
        elif ty == "T":
            return 3
        else:
            return 0

    def isAromatic(self):
        return self.__getAttribute("pdbx_aromatic_flag") != "N"

    def getCIPStereo(self):
        tS = self.__getAttribute("pdbx_stereo_config")
        tS = tS if tS not in [".", "?", "N"] else None
        if tS and tS not in ["R", "S"]:
            logger.error("Unexpected CIP stereo type %r in %r", tS, self.__rowData)
        return tS

    def getFormalCharge(self):
        try:
            return int(self.__getAttribute("charge"))
        except Exception:
            return 0

    def getFormalChargeAsInt(self):
        try:
            tch = self.getFormalCharge()
            iCharge = int(tch) if tch and tch not in [".", "?"] else 0
            return iCharge
        except Exception:
            return 0

    def hasModelCoordinates(self):
        xV, yV, zV = self.getModelCoordinates()
        # x=self.__getAttribute('model_Cartn_x')
        # y=self.__getAttribute('model_Cartn_y')
        # z=self.__getAttribute('model_Cartn_z')
        #
        return (xV is not None) and (yV is not None) and (zV is not None)

    def hasIdealCoordinates(self):
        xV, yV, zV = self.getIdealCoordinates()
        # x=self.__getAttribute('pdbx_model_Cartn_x_ideal')
        # y=self.__getAttribute('pdbx_model_Cartn_y_ideal')
        # z=self.__getAttribute('pdbx_model_Cartn_z_ideal')
        #
        return (xV is not None) and (yV is not None) and (zV is not None)

    def getModelCoordinates(self):
        """ Returns (x,y,z)
        """
        try:
            xV = float(self.__getAttribute("model_Cartn_x"))
            yV = float(self.__getAttribute("model_Cartn_y"))
            zV = float(self.__getAttribute("model_Cartn_z"))
            return (xV, yV, zV)
        except Exception:
            return (None, None, None)

    def getIdealCoordinates(self):
        """ Returns (x,y,z)
        """
        try:
            xV = float(self.__getAttribute("pdbx_model_Cartn_x_ideal"))
            yV = float(self.__getAttribute("pdbx_model_Cartn_y_ideal"))
            zV = float(self.__getAttribute("pdbx_model_Cartn_z_ideal"))
            return (xV, yV, zV)
        except Exception:
            return (None, None, None)

    def dump(self):
        logger.info(" --rowData: %r", self.__rowData)


class PdbxChemCompBondPersist(object):
    """ Accessor methods chemical component bond attributes.

    """

    def __init__(self, rowData, attributeNameList):
        self.__rowData = rowData
        self.__attributeNameList = attributeNameList

    def __getAttribute(self, name):
        try:
            i = self.__attributeNameList.index(name)
            return self.__rowData[i]
        except Exception:
            return None

    def set(self, rowData=None):
        self.__rowData = rowData
        return self

    def getBond(self):
        """ Returns (atomI,atomJ) atom ids from the atom list.
        """
        return (self.__getAttribute("atom_id_1"), self.__getAttribute("atom_id_2"))

    def getType(self):
        tS = self.__getAttribute("value_order")
        tS = tS if tS not in [".", "?"] else None
        if tS and tS not in ["SING", "DOUB", "TRIP", "QUAD"]:
            logger.error("Unexpected bond type %r in %r", tS, self.__rowData)
        return tS

    def getIntegerType(self):
        bT = self.__getAttribute("value_order")
        if bT == "SING":
            return 1
        elif bT == "DOUB":
            return 2
        elif bT == "TRIP":
            return 3
        elif bT == "QUAD":
            return 4
        else:
            return 0

    def isAromatic(self):
        return self.__getAttribute("pdbx_aromatic_flag") == "Y"

    def getStereo(self):
        tS = self.__getAttribute("pdbx_stereo_config")
        tS = tS if tS not in [".", "?", "N"] else None
        if tS and tS not in ["E", "Z"]:
            logger.error("Unexpected CIP stereo type %r in %r", tS, self.__rowData)
        return tS

    def hasStereo(self):
        return self.__getAttribute("pdbx_stereo_config") != "N"

    def dump(self):
        logger.info(" -- rowData: %r", self.__rowData)


class PdbxChemCompIdentifierPersist(object):
    """ Accessor methods chemical component identifier attributes.

    """

    def __init__(self, rowData, attributeNameList):
        self.__rowData = rowData
        self.__attributeNameList = attributeNameList

    def __getAttribute(self, name):
        try:
            i = self.__attributeNameList.index(name)
            return self.__rowData[i]
        except Exception:
            return None

    def set(self, rowData=None):
        self.__rowData = rowData
        return self

    def getIdentifier(self):
        """ Returns the value of the identifier.
        """
        return self.__getAttribute("identifier")

    def getType(self):
        return self.__getAttribute("type")

    def getProgram(self):
        return self.__getAttribute("program")

    def getProgramVersion(self):
        return self.__getAttribute("program_version")

    def dump(self):
        logger.info("-- rowData: %r", self.__rowData)


class PdbxChemCompDescriptorPersist(object):
    """ Accessor methods chemical component descriptor  attributes.

    """

    def __init__(self, rowData, attributeNameList):
        self.__rowData = rowData
        self.__attributeNameList = attributeNameList
        self.__buildTypeD = {
            ("SMILES_CANONICAL", "OPENEYE OETOOLKITS"): "oe-iso-smiles",
            ("SMILES", "OPENEYE OETOOLKITS"): "oe-smiles",
            ("SMILES", "ACDLABS"): "acdlabs-smiles",
            ("SMILES", "CACTVS"): "cactvs-smiles",
            ("SMILES_CANONICAL", "CACTVS"): "cactvs-iso-smiles",
            ("INCHI", "INCHI"): "inchi",
            ("INCHIKEY", "INCHI"): "inchikey",
        }

    def __getAttribute(self, name):
        try:
            i = self.__attributeNameList.index(name)
            return self.__rowData[i]
        except Exception:
            return None

    def set(self, rowData=None):
        self.__rowData = rowData
        return self

    def getDescriptor(self):
        """ Returns the value of the descriptor.
        """
        return self.__getAttribute("descriptor")

    def getType(self):
        return self.__getAttribute("type")

    def getProgram(self):
        return self.__getAttribute("program")

    def getProgramVersion(self):
        return self.__getAttribute("program_version")

    def getMolBuildType(self):
        bT = None
        try:
            dType = self.getType()
            dProg = self.getProgram()
            bT = self.__buildTypeD[(dType.upper(), dProg.upper())]
        except Exception as e:
            logger.error("Unknown descriptor build type %r %r %r", dType, dProg, str(e))
        return bT

    def dump(self):
        logger.info(" -- rowData: %r", self.__rowData)


class PdbxChemCompAuditPersist(object):
    """ Accessor methods chemical component audit details.

    """

    def __init__(self, rowData, attributeNameList):
        self.__rowData = rowData
        self.__attributeNameList = attributeNameList

    def __getAttribute(self, name):
        try:
            i = self.__attributeNameList.index(name)
            return self.__rowData[i]
        except Exception:
            return None

    def set(self, rowData=None):
        self.__rowData = rowData
        return self

    def getActionType(self):
        """ Returns the value of the action type.
        """
        return self.__getAttribute("action_type")

    def getDate(self):
        """ Returns the value of audit date.
        """
        return self.__getAttribute("date")

    def getProcessingSite(self):
        """ Returns the value of processing site.
        """
        return self.__getAttribute("processing_site")

    def getAnnotator(self):
        """ Returns the value of audit annotator.
        """
        return self.__getAttribute("annotator")

    def getDetails(self):
        """ Returns the value of audit details.
        """
        return self.__getAttribute("details")



##
# File:    ChemCompDepictWrapper.py
# Author:  jdw
# Date:    9-Mar-2020
# Version: 0.001
#
# Updates:
#
##
"""
Wrapper for chemical component depiction operations.
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import os
import platform
import resource
import time

from rcsb.utils.chem.ChemCompSearchWrapper import ChemCompSearchWrapper
from rcsb.utils.chem.OeDepict import OeDepict
from rcsb.utils.chem.OeDepictAlign import OeDepictMCSAlignPage
from rcsb.utils.chem.OeIoUtils import OeIoUtils
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.SingletonClass import SingletonClass

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logger = logging.getLogger(__name__)


class ChemCompDepictWrapper(SingletonClass):
    """ Wrapper for chemical component depiction operations.
    """

    def __init__(self):
        self.__startTime = time.time()
        # ---
        self.__workPath = "."
        self.__mU = MarshalUtil(workPath=self.__workPath)
        self.__configD = None
        self.__cachePath = None
        # ---
        self.__oesU = None
        # ---
        self.__statusDescriptorError = -100
        self.__searchError = -200
        self.__searchSuccess = 0
        self.__imageCount = 0

    def readConfig(self):
        #
        ok = False
        try:
            self.__cachePath = os.environ.get("CHEM_DEPICT_CACHE_PATH", ".")
            configFileName = os.environ.get("CHEM_DEPICT_CONFIG_FILE_NAME", "depict-config.json")
            configFilePath = os.path.join(self.__cachePath, "config", configFileName)
            configD = self.__mU.doImport(configFilePath, fmt="json")
            logger.debug("configD: %r", configD)
            if configD and (len(configD) >= 2) and float(configD["versionNumber"]) > 0.1:
                logger.info("Read version %r sections %r from %s", configD["versionNumber"], list(configD.keys()), configFilePath)
                ok = True
                self.__configD = configD
            else:
                logger.error("Reading config file fails from path %r", configFilePath)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ok = False
        return ok

    def setImageCount(self, imageCount):
        self.__imageCount = imageCount

    def getImageCount(self):
        return self.__imageCount

    def __makeImagePath(self):
        imageDirPath = self.__configD["imageDirPath"] if self.__configD and "imageDirPath" in self.__configD else "."
        fileRotateIncrement = self.__configD["fileRotateIncrement"] if self.__configD and "fileRotateIncrement" in self.__configD else 50
        ic = self.__imageCount % fileRotateIncrement
        imagePath = os.path.join(imageDirPath, "image-%s.svg" % ic)
        return imagePath

    def depictMolecule(self, identifier, identifierType, imagePath=None, **kwargs):
        """Create depiction from InChI, SMILES descriptors or PDB identifier.
        """
        try:
            imagePath = imagePath if imagePath else self.__makeImagePath()
            oeio = OeIoUtils()
            if identifierType.lower() in ["smiles"]:
                oeMol = oeio.smilesToMol(identifier)
            elif identifierType.lower() in ["inchi"]:
                oeMol = oeio.inchiToMol(identifier)
            elif identifierType.lower() in ["identifierpdb"]:
                ccsw = ChemCompSearchWrapper()
                oesmP = ccsw.getSearchMoleculeProvider()
                oeMol = oesmP.getMol(identifier)
            #
            ok = self.__depictOne(oeMol, imagePath, **kwargs)
            return imagePath if ok else None
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return None

    def __depictOne(self, oeMol, imagePath, **kwargs):
        """Single

        Args:
            oeMol (object): instance of an OE graph molecule
            imagePath (string): file path for image

        Returns:
            bool: True for success or False otherwise
        """
        try:
            title = kwargs.get("title", None)
            oed = OeDepict()
            oed.setMolTitleList([("Target", oeMol, title)])

            # ---
            bondDisplayWidth = 10.0
            numAtoms = oeMol.NumAtoms()
            if numAtoms > 100 and numAtoms <= 200:
                bondDisplayWidth = 6.0
            elif numAtoms > 200:
                bondDisplayWidth = 4.0
            # ---
            oed.setDisplayOptions(
                imageSizeX=kwargs.get("imageSizeX", 2500),
                imageSizeY=kwargs.get("imageSizeX", 2500),
                labelAtomName=kwargs.get("labelAtomName", False),
                labelAtomCIPStereo=kwargs.get("labelAtomCIPStereo", True),
                labelAtomIndex=kwargs.get("labelAtomIndex", False),
                labelBondIndex=kwargs.get("labelBondIndex", False),
                labelBondCIPStereo=kwargs.get("labelBondCIPStereo", True),
                cellBorders=kwargs.get("cellBorders", True),
                bondDisplayWidth=bondDisplayWidth,
            )
            oed.setGridOptions(rows=1, cols=1, cellBorders=False)
            oed.prepare()
            oed.write(imagePath)
            self.__imageCount += 1
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def status(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Status at %s (up %.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def alignMoleculePair(self, refIdentifier, refIdentifierType, fitIdentifier, fitIdentifierType, imagePath=None, **kwargs):
        """Create aligned depiction for a target molecule InChI, SMILES descriptors or PDB identifier.
        """
        try:
            imagePath = imagePath if imagePath else self.__makeImagePath()
            oeio = OeIoUtils()
            ccsw = ChemCompSearchWrapper()
            oesmP = ccsw.getSearchMoleculeProvider()
            # ---
            if refIdentifierType.lower() in ["smiles"]:
                oeMolRef = oeio.smilesToMol(refIdentifier)
            elif refIdentifierType.lower() in ["inchi"]:
                oeMolRef = oeio.inchiToMol(refIdentifier)
            elif refIdentifierType.lower() in ["identifierpdb"]:
                oeMolRef = oesmP.getMol(refIdentifier)
            #
            if fitIdentifierType.lower() in ["smiles"]:
                oeMolFit = oeio.smilesToMol(fitIdentifier)
            elif fitIdentifierType.lower() in ["inchi"]:
                oeMolFit = oeio.inchiToMol(fitIdentifier)
            elif fitIdentifierType.lower() in ["identifierpdb"]:
                oeMolFit = oesmP.getMol(fitIdentifier)
            # ---
            logger.info("oeMolRef atoms %r", oeMolRef.NumAtoms())
            logger.info("oeMolFit atoms %r", oeMolFit.NumAtoms())

            displayIdRef = "Ref"
            displayIdFit = "Fit"
            ok = self.__depictAlignedPair(oeMolRef, displayIdRef, oeMolFit, displayIdFit, imagePath, **kwargs)
            return imagePath if ok else None
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return None

    def __depictAlignedPair(self, oeMolRef, displayIdRef, oeMolFit, displayIdFit, imagePath, **kwargs):
        """ Depict pairwise MCSS alignment
        """
        try:
            #
            oed = OeDepictMCSAlignPage()
            oed.setSearchType(sType="relaxed")
            #
            oed.setRefMol(oeMolRef, displayIdRef)
            oed.setFitMol(oeMolFit, displayIdFit)
            #
            # imagePath = self.__makeImagePath()
            # ---
            bondDisplayWidth = 10.0
            numAtomsRef = oeMolRef.NumAtoms()
            if numAtomsRef > 100 and numAtomsRef <= 200:
                bondDisplayWidth = 6.0
            elif numAtomsRef > 200:
                bondDisplayWidth = 4.0
            # ---
            oed.setDisplayOptions(
                imageSizeX=kwargs.get("imageSizeX", 2500),
                imageSizeY=kwargs.get("imageSizeX", 2500),
                labelAtomName=kwargs.get("labelAtomName", False),
                labelAtomCIPStereo=kwargs.get("labelAtomCIPStereo", True),
                labelAtomIndex=kwargs.get("labelAtomIndex", False),
                labelBondIndex=kwargs.get("labelBondIndex", False),
                labelBondCIPStereo=kwargs.get("labelBondCIPStereo", True),
                cellBorders=kwargs.get("cellBorders", True),
                bondDisplayWidth=bondDisplayWidth,
                highlightStyleFit=kwargs.get("highlightStyleFit", "ballAndStickInverse"),
            )
            #
            aML = oed.alignPair(imagePath=imagePath)
            logger.info("Aligned atom count %d", len(aML))
            #
            # self.assertGreater(len(aML), 1)
            # if aML:
            #    for (rCC, rAt, tCC, tAt) in aML:
            #        logger.debug("%5s %-5s %5s %-5s", rCC, rAt, tCC, tAt)
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False
